@@ -2,13 +2,17 @@
 # Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
 # GNU Lesser General Public License version 3 (see the file LICENSE).
 
+import os
+import sys
 from mock import patch
 from unittest import TestCase
 import httplib2
 import shutil
 import tempfile
 
-from piston_mini_client import (PistonAPI, returns_json, returns,
+from piston_mini_client import (
+    APIError,
+    PistonAPI, returns_json, returns,
     returns_list_of, PistonResponseObject, PistonSerializable,
     OfflineModeException, safename)
 from piston_mini_client.auth import BasicAuthorizer
@@ -23,11 +27,11 @@ class PistonAPITestCase(TestCase):
 
     @patch('httplib2.Http.request')
     def test_request(self, mock_request):
-        mock_request.return_value = ({'status': '200'
-        }, 'hello world!')
+        mock_request.return_value = ({'status': '200'}, 'hello world!')
         api = self.CoffeeAPI()
         api._request('/foo', 'POST', body='foo=bar')
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='foo=bar', headers={}, method='POST')
 
     @patch('httplib2.Http.request')
@@ -37,7 +41,8 @@ class PistonAPITestCase(TestCase):
         tmpdir = tempfile.mkdtemp()
         http = httplib2.Http(cache=tmpdir)
         cachekey = self.CoffeeAPI.default_service_root + path
-        http.cache.set(cachekey, "header\r\n\r\nmy_cached_body\n")
+        http.cache.set(
+            cachekey, "header\r\n\r\nmy_cached_body\n".encode("utf-8"))
         # ensure that we trigger a error like when offline (no dns)
         mock_request.side_effect = httplib2.ServerNotFoundError("")
         api = self.CoffeeAPI(cachedir=tmpdir, offline_mode=True)
@@ -64,7 +69,8 @@ class PistonAPITestCase(TestCase):
         http = httplib2.Http(cache=cache)
         cachekey = self.CoffeeAPI.default_service_root + path
         http.cache.set(
-            cachekey, "header\r\n\r\nmy_cached_body_from_long_path\n")
+            cachekey.encode('utf-8'),
+            "header\r\n\r\nmy_cached_body_from_long_path\n".encode('utf-8'))
         # ensure that we trigger a error like when offline (no dns)
         mock_request.side_effect = httplib2.ServerNotFoundError("")
         api = self.CoffeeAPI(cachedir=tmpdir, offline_mode=True)
@@ -78,11 +84,11 @@ class PistonAPITestCase(TestCase):
     def test_auth_request(self, mock_request):
         mock_request.return_value = ({'status': '200'}, '""')
         api = self.CoffeeAPI(auth=BasicAuthorizer(username='foo',
-            password='bar'))
+                             password='bar'))
         api._request('/fee', 'GET')
         kwargs = mock_request.call_args[1]
         self.assertEqual(kwargs['headers']['Authorization'],
-            'Basic Zm9vOmJhcg==\n')
+                         'Basic Zm9vOmJhcg==')
         self.assertEqual(kwargs['method'], 'GET')
 
     @patch('httplib2.Http.request')
@@ -91,7 +97,7 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._post('/serve', data={'foo': 'bar'})
         kwargs = mock_request.call_args[1]
-        self.assertEqual(kwargs['headers']['Content-type'], 'application/json')
+        self.assertEqual(kwargs['headers']['Content-Type'], 'application/json')
         self.assertEqual(kwargs['method'], 'POST')
 
     @patch('httplib2.Http.request')
@@ -103,7 +109,7 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._post('/serve', data=MyCoffeeRequest(strength='mild'))
         kwargs = mock_request.call_args[1]
-        self.assertEqual(kwargs['headers']['Content-type'], 'application/json')
+        self.assertEqual(kwargs['headers']['Content-Type'], 'application/json')
         self.assertEqual(kwargs['method'], 'POST')
 
     @patch('httplib2.Http.request')
@@ -111,10 +117,10 @@ class PistonAPITestCase(TestCase):
         mock_request.return_value = ({'status': '200'}, '""')
         api = self.CoffeeAPI()
         api._post('/serve', data={'foo': 'bar'},
-            content_type='application/x-www-form-urlencoded')
+                  content_type='application/x-www-form-urlencoded')
         kwargs = mock_request.call_args[1]
-        self.assertEqual(kwargs['headers']['Content-type'],
-            'application/x-www-form-urlencoded')
+        self.assertEqual(kwargs['headers']['Content-Type'],
+                         'application/x-www-form-urlencoded')
         self.assertEqual(kwargs['method'], 'POST')
 
     @patch('httplib2.Http.request')
@@ -145,7 +151,7 @@ class PistonAPITestCase(TestCase):
             body = api._get('/simmer')
             self.assertEqual(expected_body, body)
             mock_request.assert_called_with('http://localhost:12345/simmer',
-                body='', headers={}, method='GET')
+                                            body='', headers={}, method='GET')
 
     @patch('httplib2.Http.request')
     def test_get_with_extra_args(self, mock_request):
@@ -182,7 +188,7 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._request('/foo', 'GET', scheme='https')
         mock_request.assert_called_with('https://localhost:12345/foo',
-            body='', headers={}, method='GET')
+                                        body='', headers={}, method='GET')
 
     @patch('httplib2.Http.request')
     def test_get_scheme_switch_to_https(self, mock_request):
@@ -190,15 +196,16 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._get('/foo', scheme='https')
         mock_request.assert_called_with('https://localhost:12345/foo',
-            body='', headers={}, method='GET')
+                                        body='', headers={}, method='GET')
 
     @patch('httplib2.Http.request')
     def test_post_scheme_switch_to_https(self, mock_request):
         mock_request.return_value = ({'status': '200'}, '""')
         api = self.CoffeeAPI()
         api._post('/foo', scheme='https')
-        mock_request.assert_called_with('https://localhost:12345/foo',
-            body='null', headers={'Content-type': 'application/json'},
+        mock_request.assert_called_with(
+            'https://localhost:12345/foo',
+            body='null', headers={'Content-Type': 'application/json'},
             method='POST')
 
     @patch('httplib2.Http.request')
@@ -216,7 +223,7 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._put('/serve', data={'foo': 'bar'})
         kwargs = mock_request.call_args[1]
-        self.assertEqual(kwargs['headers']['Content-type'], 'application/json')
+        self.assertEqual(kwargs['headers']['Content-Type'], 'application/json')
         self.assertEqual(kwargs['method'], 'PUT')
 
     @patch('httplib2.Http.request')
@@ -228,7 +235,7 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._put('/serve', data=MyCoffeeRequest(strength='mild'))
         kwargs = mock_request.call_args[1]
-        self.assertEqual(kwargs['headers']['Content-type'], 'application/json')
+        self.assertEqual(kwargs['headers']['Content-Type'], 'application/json')
         self.assertEqual(kwargs['method'], 'PUT')
 
     @patch('httplib2.Http.request')
@@ -245,8 +252,9 @@ class PistonAPITestCase(TestCase):
         mock_request.return_value = ({'status': '200'}, '""')
         api = self.CoffeeAPI()
         api._put('/foo', scheme='https')
-        mock_request.assert_called_with('https://localhost:12345/foo',
-            body='null', headers={'Content-type': 'application/json'},
+        mock_request.assert_called_with(
+            'https://localhost:12345/foo',
+            body='null', headers={'Content-Type': 'application/json'},
             method='PUT')
 
     @patch('httplib2.Http.request')
@@ -254,11 +262,11 @@ class PistonAPITestCase(TestCase):
         mock_request.return_value = ({'status': '200'}, '""')
         api = self.CoffeeAPI()
         api._put('/serve', data={'foo': 'bar'},
-            content_type='application/x-www-form-urlencoded')
+                 content_type='application/x-www-form-urlencoded')
         kwargs = mock_request.call_args[1]
         self.assertEqual(kwargs['body'], 'foo=bar')
-        self.assertEqual(kwargs['headers']['Content-type'],
-            'application/x-www-form-urlencoded')
+        self.assertEqual(kwargs['headers']['Content-Type'],
+                         'application/x-www-form-urlencoded')
         self.assertEqual(kwargs['method'], 'PUT')
 
     @patch('httplib2.Http.request')
@@ -268,17 +276,21 @@ class PistonAPITestCase(TestCase):
         api.extra_headers = {'X-Foo': 'bar'}
         api._get('/foo')
         expected_headers = {'X-Foo': 'bar'}
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='', headers=expected_headers, method='GET')
         api._delete('/foo')
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='', headers=expected_headers, method='DELETE')
-        expected_headers['Content-type'] = 'application/json'
+        expected_headers['Content-Type'] = 'application/json'
         api._post('/foo')
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='null', headers=expected_headers, method='POST')
         api._put('/foo')
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='null', headers=expected_headers, method='PUT')
 
     @patch('httplib2.Http.request')
@@ -287,17 +299,21 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._get('/foo', extra_headers={'X-Foo': 'bar'})
         expected_headers = {'X-Foo': 'bar'}
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='', headers=expected_headers, method='GET')
         api._delete('/foo', extra_headers={'X-Foo': 'bar'})
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='', headers=expected_headers, method='DELETE')
-        expected_headers['Content-type'] = 'application/json'
+        expected_headers['Content-Type'] = 'application/json'
         api._post('/foo', extra_headers={'X-Foo': 'bar'})
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='null', headers=expected_headers, method='POST')
         api._put('/foo', extra_headers={'X-Foo': 'bar'})
-        mock_request.assert_called_with('http://localhost:12345/foo',
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
             body='null', headers=expected_headers, method='PUT')
 
     @patch('httplib2.Http.request')
@@ -312,13 +328,15 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api.serializers = {'application/json': MySerializer()}
         api._post('/foo', data=[])
-        mock_request.assert_called_with('http://localhost:12345/foo',
-            body=expected, headers={'Content-type': 'application/json'},
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
+            body=expected, headers={'Content-Type': 'application/json'},
             method='POST')
 
         api._put('/foo', data=None)
-        mock_request.assert_called_with('http://localhost:12345/foo',
-            body=expected, headers={'Content-type': 'application/json'},
+        mock_request.assert_called_with(
+            'http://localhost:12345/foo',
+            body=expected, headers={'Content-Type': 'application/json'},
             method='PUT')
 
     @patch('httplib2.Http.request')
@@ -336,7 +354,24 @@ class PistonAPITestCase(TestCase):
         api = self.CoffeeAPI()
         api._delete('/sugar/12', scheme='https')
         mock_request.assert_called_with('https://localhost:12345/sugar/12',
-            body='', headers={}, method='DELETE')
+                                        body='', headers={}, method='DELETE')
+
+    def test_cachedir_crash_race_lp803280(self):
+        def _simulate_race(path):
+            """ this helper simulates the actual race when after
+                os.path.exists() a different process creates the dir
+            """
+            patcher.stop()
+            os.makedirs(path)
+        # this simulates a race when multiple piston-mini-client helpers
+        # try to create the cachedir at the same time (LP: #803280)
+        patcher = patch('os.path.exists')
+        mock = patcher.start()
+        mock.return_value = False
+        mock.side_effect = _simulate_race
+        tmpdir = os.path.join(tempfile.mkdtemp(), "foo")
+        api = self.CoffeeAPI(cachedir=tmpdir)
+        self.assertNotEqual(api, None)
 
 
 class PistonResponseObjectTestCase(TestCase):
@@ -361,6 +396,30 @@ class ReturnsJSONTestCase(TestCase):
         result = MyAPI().func()
         self.assertEqual({"foo": "bar", "baz": 42}, result)
 
+    def test_returns_json_error(self):
+        # If the method does not return JSON, we raise a ValueError saying
+        # what it did return.
+        not_json = 'This cannot be valid JSON'
+
+        class MyAPI(PistonAPI):
+            default_service_root = 'http://foo'
+
+            @returns_json
+            def its_a_lie(self):
+                return not_json
+
+        # If we used unittest2 or testtools, we wouldn't need this.
+        try:
+            result = MyAPI().its_a_lie()
+        except APIError:
+            # This is the Python 2 & 3 compatible way of getting an exception.
+            exc_type, exc_value, tb = sys.exc_info()
+            self.assertEqual(not_json, exc_value.body)
+        else:
+            self.fail(
+                "MyAPI().its_a_lie did not raise. Returned %r instead"
+                % (result,))
+
 
 class ReturnsTestCase(TestCase):
     def test_returns(self):
